import argparse
import contextlib
import dataclasses
import inspect
import test
import textwrap
from dataclasses import dataclass, fields
from enum import Enum
from pathlib import Path
from typing import *

import pytest

import simple_parsing
from simple_parsing import ArgumentParser
from simple_parsing.helpers import subparsers

from .testutils import TestSetup, xfail, raises


@dataclass
class TrainOptions:
    """ Training Options """
    lr: float = 1e-3
    train_path: Path = Path("./train")

@dataclass
class ValidOptions:
    """ Validation Options """
    test_path: Path = Path("./test")
    metric: str = "accuracy"

@dataclass
class GlobalOptions(TestSetup):
    """ Global Options """
    # mode, either Train or Valid.
    mode: Union[TrainOptions, ValidOptions] = subparsers({
        "train": TrainOptions,
        "valid": ValidOptions,
    })
    global_arg: str = "something"

# options = GlobalOptions.setup("--global_arg 123 --help")
# print(options)
# exit()

def test_required_subparser():
    options = GlobalOptions.setup("train --lr 0.1 --train_path ./bob")
    assert isinstance(options.mode, TrainOptions)
    assert options.mode.lr == 0.1
    assert options.mode.train_path == Path("./bob")

    options = GlobalOptions.setup("valid --metric some_metric --test_path ./john")
    assert isinstance(options.mode, ValidOptions)
    assert options.mode.metric == "some_metric"
    assert options.mode.test_path == Path("./john")

def test_help_text_works():
    from contextlib import suppress
    with suppress(SystemExit):
        GlobalOptions.setup("--help")
        GlobalOptions.setup("train --help")

@dataclass
class Start:
    """Start command"""
    value: str = "start command value"

    def execute(self, verbose=False):
        print(f"Start (verbose: {verbose})")
        return self.value

@dataclass
class Stop:
    """Stop command"""
    value: str = "stop commmand value"

    def execute(self, verbose=False):
        print(f"Stop (verbose: {verbose})")
        return self.value

@dataclass
class Push:
    """Example of a subcommand."""
    subcommand: Union[Start, Stop]
    some_flag_1: str = "command1"

    def execute(self, verbose=False):
        print(f"Push (verbose: {verbose})")
        return self.subcommand.execute(verbose=verbose)


@dataclass
class Pull:
    """Other Example of a subcommand."""
    subcommand: Union[Start, Stop]
    some_flag_2: str = "command2"

    def execute(self, verbose=False):
        print(f"Pull (verbose: {verbose})")
        return self.subcommand.execute(verbose=verbose)


@dataclass
class Program(TestSetup):
    """Some top-level command"""
    command: Union[Push, Pull]
    verbose: bool = False

    def execute(self):
        print(f"Program (verbose: {self.verbose})")
        return self.command.execute(verbose=self.verbose)


def test_command_tree():
    prog: Program = Program.setup("push start --value jack")
    assert isinstance(prog.command, Push)
    assert isinstance(prog.command.subcommand, Start)
    assert prog.command.subcommand.value == "jack"
    assert prog.execute() == "jack"


def test_experiments():
    from abc import ABC
    
    @dataclass
    class Experiment(ABC):
        dataset: str
        iid: bool = True

    @dataclass
    class Mnist(Experiment):
        dataset: str = "mnist"
        iid: bool = True
    
    @dataclass
    class MnistContinual(Experiment):
        dataset: str = "mnist"
        iid: bool = False

    @dataclass
    class Config:
        experiment: Experiment = subparsers({
            "mnist": Mnist,
            "mnist_continual": MnistContinual,
        })

    for field in dataclasses.fields(Config):
        assert simple_parsing.utils.is_subparser_field(field), field

    parser = ArgumentParser()
    parser.add_arguments(Config, "config")
    
    with raises(SystemExit):
        args = parser.parse_args()
    
    args = parser.parse_args("mnist".split())
    experiment = args.config.experiment
    assert isinstance(experiment, Mnist)
    assert experiment.dataset == "mnist"
    assert experiment.iid == True

    args = parser.parse_args("mnist_continual".split())
    experiment = args.config.experiment
    assert isinstance(experiment, MnistContinual)
    assert experiment.dataset == "mnist"
    assert experiment.iid == False

    

if __name__ == "__main__":
    import sys
    print("ARGS:", " ".join(sys.argv[1:]))
    prog: Program = Program.setup(" ".join(sys.argv[1:]))
    print(prog)
    print(prog.execute())
    exit()



from dataclasses import dataclass
from typing import Optional
from simple_parsing import ArgumentParser



@dataclass
class Config:
    seed: Optional[int] = None

def test_optional_seed():
    """Test that a value marked as Optional works fine.
    
    (Reproduces https://github.com/lebrice/SimpleParsing/issues/14#issue-562538623)
    """
    parser = ArgumentParser()
    parser.add_arguments(Config, dest="config")

    args = parser.parse_args("".split())
    config: Config = args.config
    assert config == Config()

    args = parser.parse_args("--seed 123".split())
    config: Config = args.config
    assert config == Config(123)


import copy
import inspect
import json
import warnings
from collections import OrderedDict
from dataclasses import Field, asdict, dataclass, fields, is_dataclass
from functools import singledispatch
from pathlib import Path
from typing import *
from typing import IO, TypeVar

import typing_inspect as tpi

from ...logging_utils import get_logger
from ...utils import get_type_arguments, is_dict, is_list, is_union
from .decoding import decoding_fns, register_decoding_fn
from .encoding import SimpleJsonEncoder, encode

logger = get_logger(__file__)

Dataclass = TypeVar("Dataclass")
D = TypeVar("D", bound="Serializable")

try:
    import yaml
    def ordered_dict_constructor(loader: yaml.Loader, node: yaml.Node):
        value = loader.construct_sequence(node)
        return OrderedDict(*value)

    def ordered_dict_representer(dumper: yaml.Dumper, instance: OrderedDict) -> yaml.Node:
        node = dumper.represent_sequence("OrderedDict", instance.items())
        return node

    yaml.add_representer(OrderedDict, ordered_dict_representer)
    yaml.add_constructor("OrderedDict", ordered_dict_constructor)
    yaml.add_constructor("tag:yaml.org,2002:python/object/apply:collections.OrderedDict", ordered_dict_constructor)

except ImportError:
    pass


@dataclass
class Serializable:
    """Makes a dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ... 
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {"a": 123, "b": 456}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    Config(a=123, b='456')
    >>> assert config == config_
    """
    subclasses: ClassVar[List[Type[D]]] = []
    decode_into_subclasses: ClassVar[bool] = False

    def __init_subclass__(cls, decode_into_subclasses: bool=None, add_variants: bool=True):
        logger.debug(f"Registering a new Serializable subclass: {cls}")
        super().__init_subclass__()
        if decode_into_subclasses is None:
            # if decode_into_subclasses is None, we will use the value of the
            # parent class, if it is also a subclass of Serializable.
            # Skip the class itself as well as object.
            parents = cls.mro()[1:-1] 
            logger.debug(f"parents: {parents}")

            for parent in parents:
                if parent in Serializable.subclasses and parent is not Serializable:
                    decode_into_subclasses = parent.decode_into_subclasses
                    logger.debug(f"Parent class {parent} has decode_into_subclasses = {decode_into_subclasses}")
                    break

        cls.decode_into_subclasses = decode_into_subclasses or False
        if cls not in Serializable.subclasses:
            Serializable.subclasses.append(cls)

        encode.register(cls, cls.to_dict)
        register_decoding_fn(cls, cls.from_dict, add_variants=add_variants)

    def to_dict(self, dict_factory:Type[Dict]=dict, recurse: bool=True) -> Dict:
        """ Serializes this dataclass to a dict.
        
        NOTE: This 'extends' the `asdict()` function from
        the `dataclasses` package, allowing us to not include some fields in the
        dict, or to perform some kind of custom encoding (for instance,
        detaching `Tensor` objects before serializing the dataclass to a dict).
        """
        d: Dict[str, Any] = dict_factory()
        for f in fields(self):
            name = f.name
            value = getattr(self, name)
            T = f.type

            # Do not include in dict if some corresponding flag was set in metadata.
            include_in_dict = f.metadata.get("to_dict", True)
            if not include_in_dict:
                continue

            custom_encoding_fn = f.metadata.get("encoding_fn")
            if custom_encoding_fn:
                # Use a custom encoding function if there is one.
                d[name] = custom_encoding_fn(value)
                continue

            encoding_fn = encode
            if isinstance(value, Serializable) and recurse:
                encoded = value.to_dict(dict_factory=dict_factory, recurse=recurse)
                logger.debug(f"Encoded Serializable field {name}: {encoded}")
            else:
                try:
                    encoded = encoding_fn(value)
                except Exception as e:
                    logger.error(f"Unable to encode value {value} of type {type(value)}! Leaving it as-is. (exception: {e})")
                    encoded = value
            d[name] = encoded
        return d

    @classmethod
    def from_dict(cls: Type[D], obj: Dict, drop_extra_fields: bool=None) -> D:
        """ Parses an instance of `cls` from the given dict.
        
        NOTE: If the `decode_into_subclasses` class attribute is set to True (or
        if `decode_into_subclasses=True` was passed in the class definition),
        then if there are keys in the dict that aren't fields of the dataclass,
        this will decode the dict into an instance the first subclass of `cls`
        which has all required field names present in the dictionary.
        
        Passing `drop_extra_fields=None` (default) will use the class attribute
        described above.
        Passing `drop_extra_fields=True` will decode the dict into an instance
        of `cls` and drop the extra keys in the dict.
        Passing `drop_extra_fields=False` forces the above-mentioned behaviour.
        """
        if drop_extra_fields is None:
            drop_extra_fields = not cls.decode_into_subclasses
        return from_dict(cls, obj, drop_extra_fields=drop_extra_fields)

    def dump(self, fp: IO[str], dump_fn=json.dump, **kwargs) -> None:
        # Convert `self` into a dict.
        d = self.to_dict()
        # Serialize that dict to the file, using dump_fn.
        dump_fn(d, fp, **kwargs)
    
    def dump_json(self, fp: IO[str], dump_fn=json.dump, **kwargs) -> str:
        return self.dump(fp, dump_fn=dump_fn, **kwargs)

    def dump_yaml(self, fp: IO[str], dump_fn=None, **kwargs) -> str:
        import yaml
        if dump_fn is None:
            dump_fn = yaml.dump
        return self.dump(fp, dump_fn=dump_fn, **kwargs)

    def dumps(self, dump_fn=json.dumps, **kwargs) -> str:
        d = self.to_dict()
        return dump_fn(d, **kwargs)

    def dumps_json(self, dump_fn=json.dumps, **kwargs) -> str:
        kwargs.setdefault("cls", SimpleJsonEncoder)
        return self.dumps(dump_fn=dump_fn, **kwargs)

    def dumps_yaml(self, dump_fn=None, **kwargs) -> str:
        import yaml
        if dump_fn is None:
            dump_fn = yaml.dump
        return self.dumps(dump_fn=dump_fn, **kwargs)

    @classmethod
    def load(cls: Type[D], path: Union[Path, str, IO[str]], drop_extra_fields: bool=None, load_fn=None, **kwargs) -> D:
        """Loads an instance of `cls` from the given file.
        
        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[Path, str, IO[str]]): Path or Path string or open file.
            drop_extra_fields (bool, optional): Wether to drop extra fields or 
                to decode the dictionary into the first subclass with matching
                fields. Defaults to None, in which case we use the value of
                `cls.decode_into_subclasses`. 
                For more info, see `cls.from_dict`.
            load_fn ([type], optional): Which loading function to use. Defaults
                to None, in which case we try to use the appropriate loading
                function depending on `path.suffix`:
                {
                    ".yml": yaml.full_load,
                    ".yaml": yaml.full_load,
                    ".json": json.load,
                    ".pth": torch.load,
                    ".pkl": pickle.load,
                }

        Raises:
            RuntimeError: If the extension of `path` is unsupported.

        Returns:
            D: An instance of `cls`.
        """
        if isinstance(path, str):
            path = Path(path)

        if load_fn is None and isinstance(path, Path):
            if path.name.endswith((".yml", ".yaml")):
                return cls.load_yaml(path, drop_extra_fields=drop_extra_fields, **kwargs)
            elif path.name.endswith(".json"):
                return cls.load_json(path, drop_extra_fields=drop_extra_fields, **kwargs)
            elif path.name.endswith(".pth"):
                import torch
                load_fn = torch.loads
            elif path.name.endswith(".npy"):
                import numpy as np
                load_fn = np.load
            elif path.name.endswith(".pkl"):
                import pickle
                load_fn = pickle.load
            warnings.warn(RuntimeWarning(
                f"Not sure how to deserialize contents of {path} to a dict, as no "
                f" load_fn was passed explicitly. Will try to use {load_fn} as the "
                f"load function, based on the path name."    
            )) 

        if load_fn is None:
            raise RuntimeError(
                f"Unable to determine what function to use in order to load "
                f"path {path} into a dictionary, since no load_fn was passed, "
                f"and the path doesn't have an unfamiliar extension.."
            )

        if isinstance(path, Path):
            path = path.open()
        return cls._load(path, load_fn=load_fn, drop_extra_fields=drop_extra_fields, **kwargs)

    @classmethod
    def _load(cls: Type[D], fp: IO[str], drop_extra_fields: bool=None, load_fn=json.load, **kwargs) -> D:
        # Load a dict from the file.
        d = load_fn(fp, **kwargs)
        # Convert the dict into an instance of the class.
        return cls.from_dict(d, drop_extra_fields=drop_extra_fields)

    @classmethod
    def load_json(cls: Type[D], path: Union[str, Path], drop_extra_fields: bool=None, load_fn=json.load, **kwargs) -> D:
        """Loads an instance from the corresponding json-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a json-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to json.load.

        Returns:
            D: an instance of the dataclass.
        """
        return cls.load(path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def load_yaml(cls: Type[D], path: Union[str, Path], drop_extra_fields: bool=None, load_fn=None, **kwargs) -> D:
        """Loads an instance from the corresponding yaml-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a yaml-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to
                None, in which case `yaml.full_load` is used.

        Returns:
            D: an instance of the dataclass.
        """
        import yaml
        if load_fn is None:
            load_fn = yaml.full_load
        return cls.load(path, load_fn=load_fn, drop_extra_fields=drop_extra_fields, **kwargs)

    def save(self, path: Union[str, Path], dump_fn=None, **kwargs) -> None:
        if not isinstance(path, Path):
            path = Path(path)

        if dump_fn is None and isinstance(path, Path):
            if path.name.endswith((".yml", ".yaml")):
                return self.save_yaml(path, **kwargs)
            elif path.name.endswith(".json"):
                return self.save_json(path, **kwargs)
            elif path.name.endswith(".pth"):
                import torch
                dump_fn = torch.save
            elif path.name.endswith(".npy"):
                import numpy as np
                dump_fn = np.save
            elif path.name.endswith(".pkl"):
                import pickle
                dump_fn = pickle.dump
            warnings.warn(RuntimeWarning(
                f"Not 100% sure how to deserialize contents of {path} to a "
                f"file as no dump_fn was passed explicitly. Will try to use "
                f"{dump_fn} as the serialization function, based on the path "
                f"suffix. ({path.suffix})" 
            ))

        if dump_fn is None:
            raise RuntimeError(
                f"Unable to determine what function to use in order to dump "
                f"path {path} into a dictionary, since no dump_fn was passed, "
                f"and the path doesn't have an unfamiliar extension: "
                f"({path.suffix})"
            )
        self._save(path, dump_fn=dump_fn, **kwargs)

    def _save(self, path: Union[str, Path], dump_fn=json.dump, **kwargs) -> None:    
        d = self.to_dict()
        logger.debug(f"saving to path {path}")
        with open(path, "w") as fp:
            dump_fn(d, fp, **kwargs)

    def save_yaml(self, path: Union[str, Path], dump_fn=None, **kwargs) -> None:
        import yaml
        if dump_fn is None:
            dump_fn = yaml.dump
        self.save(path, dump_fn=dump_fn, **kwargs)

    def save_json(self, path: Union[str, Path], dump_fn=json.dump, **kwargs) -> None:
        self.save(path, dump_fn=dump_fn, **kwargs)


    @classmethod
    def loads(cls: Type[D], s: str, drop_extra_fields: bool=None, load_fn=json.loads, **kwargs) -> D:
        d = load_fn(s, **kwargs)
        return cls.from_dict(d, drop_extra_fields=drop_extra_fields)

    @classmethod
    def loads_json(cls: Type[D], s: str, drop_extra_fields: bool=None, load_fn=json.loads, **kwargs) -> D:
        return cls.loads(s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def loads_yaml(cls: Type[D], s: str, drop_extra_fields: bool=None, load_fn=None, **kwargs) -> D:
        import yaml
        if load_fn is None:
            load_fn = yaml.full_load
        return cls.loads(s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)


@dataclass
class SimpleSerializable(Serializable, decode_into_subclasses=True):
    pass


def get_dataclass_type_from_forward_ref(forward_ref: Type, Serializable=Serializable) -> Optional[Type]:
    arg = tpi.get_forward_arg(forward_ref)
    potential_classes: List[Type] = []

    for serializable_class in Serializable.subclasses:
        if serializable_class.__name__ == arg:
            potential_classes.append(serializable_class)

    if not potential_classes:
        logger.warning(
            f"Unable to find a corresponding type for forward ref "
            f"{forward_ref} inside the registered {Serializable} subclasses. "
            f"(Consider adding {Serializable} as a base class to <{arg}>? )."
        )
        return None
    elif len(potential_classes) > 1:
        logger.warning(
            f"More than one potential {Serializable} subclass was found for "
            f"forward ref '{forward_ref}'. The appropriate dataclass will be "
            f"selected based on the matching fields. \n"
            f"Potential classes: {potential_classes}"
        )
        return Serializable
    else:
        assert len(potential_classes) == 1
        return potential_classes[0]


def get_actual_type(field_type: Type) -> Type:
    if is_union(field_type):
        logger.debug(f"field has union type: {field_type}")
        t = get_first_non_None_type(field_type)
        logger.debug(f"First non-none type: {t}")
        if t is not None:
            field_type = t

    if tpi.is_forward_ref(field_type):
        logger.debug(f"field_type {field_type} is a forward ref.")
        dc = get_dataclass_type_from_forward_ref(field_type)
        logger.debug(f"Found the corresponding type: {dc}")
        if dc is not None:
            field_type = dc
    return field_type


def decode_field(field: Field, field_value: Any, drop_extra_fields: bool=None) -> Any:
    name = field.name
    field_type = field.type
    logger.debug(f"name = {name}, field_type = {field_type} drop_extra_fields is {drop_extra_fields}")

    # If the user set a custom decoding function, we use it.
    custom_decoding_fn = field.metadata.get("decoding_fn")
    if custom_decoding_fn is not None:
        return custom_decoding_fn(field_value)

    if field_type in {str, int, bool, float}:
        return field_type(field_value)

    if field_type in decoding_fns:
        logger.debug(f"We have a decoding function for the type {field_type}")
        decoding_function = decoding_fns[field_type]
        return decoding_function(field_value)
    
    field_type = get_actual_type(field_type)
    logger.debug(f"Actual type: {field_type}")

    if is_dataclass(field_type):
        return from_dict(field_type, field_value, drop_extra_fields)
    
    elif is_list(field_type):
        item_type = get_list_item_type(field_type)
        logger.debug(f"{field_type} is a List[{item_type}]")

        if item_type and is_dataclass(item_type):    
            new_field_list_value: List = []
            for item_args in field_value:
                # Parse an instance from the dict.
                if isinstance(item_args, dict):
                    item_instance = from_dict(item_type, item_args, drop_extra_fields)
                    new_field_list_value.append(item_instance)
                else:
                    new_field_list_value.append(item_args)
            return new_field_list_value

    elif is_dict(field_type):
        key_type, value_type = get_key_and_value_types(field_type)
        logger.debug(f"Field {name} ({field_type}) is a Dict[{key_type}, {value_type}]")
        if value_type:
            new_field_dict_value: Dict = {}
            logger.debug(f"field {name} value: {field_value}")

            if isinstance(field_value, list):
                # Special case for Dict annotation, but OrderedDict value.
                field_value = OrderedDict(field_value)
                new_field_dict_value = OrderedDict()

            for k, item_args in field_value.items():
                # Parse an instance from the dict.
                if isinstance(item_args, dict):
                    item_instance = from_dict(value_type, item_args, drop_extra_fields)
                    new_field_dict_value[k] = item_instance
                else:
                    new_field_dict_value[k] = item_args
            return new_field_dict_value
    
    return field_value


def from_dict(cls: Type[Dataclass], d: Dict[str, Any], drop_extra_fields: bool=None) -> Dataclass:
    if d is None:
        return None
    obj_dict: Dict[str, Any] = d.copy()

    init_args: Dict[str, Any] = {}
    non_init_args: Dict[str, Any] = {}
        
    if drop_extra_fields is None:
        decode_into_subclasses = getattr(cls, "decode_into_subclasses", False)
        drop_extra_fields = not decode_into_subclasses
        logger.debug(f"drop_extra_fields is None, using the value of {drop_extra_fields} from the attribute on class {cls}")
        if cls is Serializable:
            logger.debug(f"The class that was passed is `Serializable`, which "
                         f"means that we should set drop_extra_fields to False.")
            drop_extra_fields = False

    logger.debug(f"from_dict called with cls {cls}, drop extra fields: {drop_extra_fields}")

    for field in fields(cls):
        name = field.name
        field_type = field.type
               
        if name not in obj_dict:
            logger.warning(f"Couldn't find the field '{name}' in the dict with keys {d.keys()}")
            continue

        field_value = obj_dict.pop(name)
        field_value = decode_field(field, field_value)

        if field.init:
            init_args[name] = field_value
        else:
            non_init_args[name] = field_value
    
    extra_args = obj_dict
    
    if extra_args:
        if drop_extra_fields:
            logger.warning(f"Dropping extra args {extra_args}")
            extra_args.clear()
        
        elif issubclass(cls, Serializable):
            # Use the first Serializable derived class that has all the required fields.
            logger.debug(f"Missing field names: {extra_args.keys()}")  

            # Find all the "registered" subclasses of `cls`. (from Serializable)
            derived_classes: List[Type] = []
            for subclass in Serializable.subclasses:
                if issubclass(subclass, cls) and subclass is not cls:
                    derived_classes.append(subclass)
            logger.debug(f"All serializable derived classes of {cls} available: {derived_classes}")

            from itertools import chain
            # All the arguments that the dataclass should be able to accept in its 'init'.
            req_init_field_names = set(chain(extra_args, init_args))
            
            # Sort the derived classes by their number of init fields, so we choose the first one that fits.
            derived_classes.sort(key=lambda dc: len(get_init_fields(dc)))

            for child_class in derived_classes:
                logger.debug(f"class name: {child_class.__name__}, mro: {child_class.mro()}") 
                child_init_fields: Dict[str, Field] = get_init_fields(child_class)
                child_init_field_names = set(child_init_fields.keys())
                
                if child_init_field_names >= req_init_field_names:
                    logger.debug(f"Using child class {child_class} instead of {cls}, since it has all the required fields.")
                    return from_dict(child_class, d, drop_extra_fields=False)
        else:
            raise RuntimeError(f"Unexpected arguments for class {cls}: {extra_args}")
    
    init_args.update(extra_args)
    try:
        instance = cls(**init_args)  # type: ignore
    except TypeError as e:
        # raise RuntimeError(f"Couldn't instantiate class {cls} using init args {init_args}.")
        raise RuntimeError(f"Couldn't instantiate class {cls} using init args {init_args.keys()}: {e}")

    for name, value in non_init_args.items():
        logger.debug(f"Setting non-init field '{name}' on the instance.")
        setattr(instance, name, value)
    return instance


def get_key_and_value_types(dict_type: Type[Dict], Serializable=Serializable) -> Tuple[Optional[Type], Optional[Type]]:
    args = get_type_arguments(dict_type)

    if len(args) != 2:
        logger.debug(f"Weird.. the type {dict_type} doesn't have 2 args: {args}")
        return None, None
    K_ = args[0]
    V_ = args[1]
    # Get rid of Unions or ForwardRefs or Optionals
    V_ = get_actual_type(V_)

    logger.debug(f"K_: {K_}, V_: {V_}")
    if isinstance(V_, tuple):
        V_ = get_first_non_None_type(V_)
    elif tpi.is_optional_type(V_):
        logger.debug(f"V_ is optional: {V_}")
        V_ = get_first_non_None_type(V_)
    return K_, V_


def get_list_item_type(list_type: Type[List]) -> Optional[Type]:
    logger.debug(f"list type: {list_type}")
    args = tpi.get_args(list_type)
    logger.debug(f"args = {args}")
    
    if not args:
        return None

    assert isinstance(args, tuple), args
    if isinstance(args[0], tuple):
        args = args[0]
    assert isinstance(args, tuple), args
    logger.debug(f"args tuple: {args}")
    V_ = get_first_non_None_type(args)
    logger.debug(f"item type: {V_}")
    assert V_ is not None
    V_ = get_actual_type(V_)

    assert not isinstance(V_, tuple), V_
    return V_


def get_init_fields(dataclass: Type) -> Dict[str, Field]:
    result: Dict[str, Field] = {}
    for field in fields(dataclass):
        if field.init:
            result[field.name] = field
    return result


def get_first_non_None_type(optional_type: Union[Type, Tuple[Type, ...]]) -> Optional[Type]:
    if not isinstance(optional_type, tuple):
        optional_type = tpi.get_args(optional_type)
    for arg in optional_type:
        if arg is not Union and arg is not type(None):
            logger.debug(f"arg: {arg} is not union? {arg is not Union}")
            logger.debug(f"arg is not type(None)? {arg is not type(None)}")
            return arg
    return None


def is_dataclass_or_optional_dataclass_type(t: Type) -> bool:
    """ Returns wether `t` is a dataclass type or an Optional[<dataclass type>].
    """
    return is_dataclass(t) or (tpi.is_optional_type(t) and
        is_dataclass(tpi.get_args(t)[0])
    )



import inspect
# from typing import ForwardRef  # type: ignore
import itertools
import warnings
from collections import OrderedDict
from dataclasses import Field, fields, is_dataclass
from typing import *

from ...logging_utils import get_logger

logger = get_logger(__file__)
# logger.setLevel(logging.DEBUG)
T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

Dataclass = TypeVar("Dataclass")

# Dictionary mapping from types/type annotations to their decoding functions.
decoding_fns: Dict[Type, Callable[[Any], Any]] = {}


def decode_optional(t: Type[T]) -> Callable[[Optional[Any]], Optional[T]]:
    decode = _get_decoding_fn(t)
    def _decode_optional(val: Optional[Any]) -> Optional[T]:
        return val if val is None else decode(val)
    return _decode_optional


def decode_list(t: Type[T]) -> Callable[[List[Any]], List[T]]:
    decode_item = _get_decoding_fn(t)
    def _decode_list(val: List[Any]) -> List[T]:
        return [decode_item(v) for v in val]
    return _decode_list


def decode_tuple(*tuple_item_types: Type[T]) -> Callable[[List[T]], Tuple[T, ...]]:
    # Get the decoding function for each item type
    decoding_fns = [
        _get_decoding_fn(t) for t in tuple_item_types
    ]
    def _decode_tuple(val: Tuple[Any, ...]) -> Tuple[T, ...]:
        return tuple(
            decoding_fns[i](v) for i, v in enumerate(val)
        )
    return _decode_tuple


def decode_set(t: Type[T]) -> Callable[[List[T]], Set[T]]:
    _decode = decode_list(t)
    def _decode_set(val: List[Any]) -> Set[T]:
        l = _decode(val)
        return set(l)
    return _decode_set


def decode_dict(K_: Type[K], V_: Type[V]) -> Callable[[List[Tuple[Any, Any]]], Dict[K, V]]:
    decode_k = _get_decoding_fn(K_)
    decode_v = _get_decoding_fn(V_) 
    def _decode_dict(val: Union[Dict[Any, Any], List[Tuple[Any, Any]]]) -> Dict[K, V]:    
        result: Dict[K, V] = {}
        if isinstance(val, list):
            result = OrderedDict()
            items = val
        else:
            items = val.items()
        for k, v in items:
            k_ = decode_k(k)
            v_ = decode_v(v)
            result[k_] = v_
        return result
    return _decode_dict


def no_op(v):
    return v

def try_constructor(t: Type[T]) -> Callable[[Any], Union[T, Any]]:
    def try_parse(val: Any) -> Union[T, Any]:
        try:
            return t(val)  # type: ignore
        except Exception as e:
            logger.error(f"Couldn't parse value {val} into an instance of type {t} using the type as a constructor: {e}")
            return val
    return try_parse

def _get_decoding_fn(t: Type[T]) -> Callable[[Any], T]:
    if t in decoding_fns:
        return decoding_fns[t]
    elif t in {int, str, float, bool}:
        return t
    warnings.warn(UserWarning(f"Unable to find a decoding function for type {t}. Will try to use the type as a constructor."))
    return try_constructor(t)


def _register(t: Type, func: Callable) -> None:
    if t not in decoding_fns:
        # logger.debug(f"Registering the type {t} with decoding function {func}")
        decoding_fns[t] = func


def register_decoding_fn(some_type: Type[T], function: Callable[[Any], T], add_variants: bool=True) -> None:
    """Register a decoding function for the type `some_type`.
    
    If `add_variants` is `True`, then also adds variants for the types:
    - Optional[some_type]
    - Optional[List[some_type]]
    - List[some_type]
    - List[Optional[some_type]]
    - List[List[some_type]]
    - List[List[Optional[some_type]]]
    - Dict[int, some_type]
    - Dict[float, some_type]
    - Dict[str, some_type]
    - Dict[bool, some_type]

    If the type is a subclass of collections.abc.Hashable, then we also add:
    - Dict[some_type, int]
    - Dict[some_type, float]
    - Dict[some_type, str]
    - Dict[some_type, bool]
    - Dict[some_type, Optional[int]]
    - Dict[some_type, Optional[float]]
    - Dict[some_type, Optional[str]]
    - Dict[some_type, Optional[bool]]
    - Set[some_type]


    NOTE: `Dict[<k>, <any of the above>]` should also be supported given how
    `from_dict` is implemented below, but I didn't test out every combination.

    #TODO: Could maybe use a different approach, where we seek out the right
    # decoding function in a 'top-down' approach. So for instance, if we
    # encounter a List[<T>], we check for <T>. if T is Dict[<K>, <V>], we check
    # for <K>, then <V>, recursively, etc. The benefit is that we wouldn't have
    # to manually create the decoding functions for all the variants. 
    """
    _register(some_type, function)

    if add_variants:
        _register(Optional[some_type],  decode_optional(some_type))
        _register(List[some_type],      decode_list(some_type))  # type: ignore
        _register(List[some_type],      decode_list(some_type))  # type: ignore
        # _register(ForwardRef(some_type.__name__), decoding_fns[some_type])

        variants: List[Type] = [
            some_type,
            Optional[some_type],
            List[some_type],  # type: ignore
            List[Optional[some_type]],  # type: ignore
            Optional[List[some_type]],  # type: ignore
            # ForwardRef(some_type.__name__),
        ]

        for item_type in variants:
            # Register decoders for Optional[<variant>]
            _register(Optional[item_type], decode_optional(item_type))  # type: ignore
            # Register decoders for List[<variant>]
            _register(List[item_type], decode_list(item_type))  # type: ignore

        key_types: List[Type] = [int, str, float, bool]
        # Register decoders for Dict<K>, <variant>]
        for key_type, value_type in itertools.product(key_types, variants):
            logger.debug(f"Registering K: {key_type}, V: {value_type}")
            decoding_fn = decode_dict(key_type, value_type)
            _register(Dict[key_type, value_type], decoding_fn)  # type: ignore
            _register(Mapping[key_type, value_type], decoding_fn)  # type: ignore
            _register(MutableMapping[key_type, value_type], decoding_fn)  # type: ignore
        
        if issubclass(some_type, Hashable):
            logger.debug(f"Adding variants for Hashable type {some_type}")
            value_types: List[Type] = [int, str, float, bool]
            # Register decoders for Dict<K>, <variant>]
            for value_type in value_types:
                logger.debug(f"Registering K: {some_type}, V: {value_type}")
                decoding_fn = decode_dict(some_type, value_type)
                _register(Dict[some_type, value_type], decoding_fn)  # type: ignore
                _register(Mapping[some_type, value_type], decoding_fn)  # type: ignore
                _register(MutableMapping[some_type, value_type], decoding_fn)  # type: ignore

            _register(Set[some_type], decode_set(some_type))  # type: ignore



import pytest

from dataclasses import dataclass
from simple_parsing import ArgumentParser, choice
from typing import Union
from .testutils import *

@dataclass
class A(TestSetup):
    color: str = choice("red", "green", "blue", default="red")



def test_choice_default():
    a = A.setup("")
    assert a.color == "red"


def test_value_not_in_choices_throws_error():
    with raises(argparse.ArgumentError):
        a = A.setup("--color orange")

def test_passed_value_works_fine():
    a = A.setup("--color red")
    assert a.color == "red"

    a = A.setup("--color green")
    assert a.color == "green"

    a = A.setup("--color blue")
    assert a.color == "blue"



@dataclass
class Base:
    value: str = "hello base"

@dataclass
class AA(Base):
    value: str = "hello a"

@dataclass
class BB(Base):
    value: str = "hello b"


def test_choice_with_dict():
    @dataclass
    class C(TestSetup):
        option: Union[AA, BB] = choice({
            "a": AA("aaa"),
            "b": BB("bbb"),
            "bob": AA("bobobo"),
        }, default="a")
    c = C.setup("--option a")
    assert c.option.value == "aaa"

    c = C.setup("--option bob")
    assert c.option.value == "bobobo"

def test_choice_with_dict_weird():
    @dataclass
    class D(TestSetup):
        option: List[Base] = choice({
            "a": [AA("aa1"), AA("aa2")],
            "b": 1.23,
            "bob": BB("bobobo"),
        }, default="a")
    c = D.setup("--option a")
    assert c.option == [AA("aa1"), AA("aa2")]

    c = D.setup("--option b")
    assert c.option == 1.23
    
    c = D.setup("--option bob")
    assert c.option == BB("bobobo")


def test_choice_with_default_instance():
    @dataclass
    class D(TestSetup):
        option: Union[AA, BB, float] = choice({
            "a": [AA("aa1"), AA("aa2")],
            "b": 1.23,
            "bob": BB("bobobo"),
        }, default="a")

    @dataclass
    class Parent(TestSetup):
        d: D = D(option=AA("parent"))


    p = Parent.setup("")
    assert p.d.option == AA("parent")



""" Utility functions that simplify defining field of dataclasses. 
"""
import argparse
import dataclasses
import functools
import json
import warnings
from dataclasses import _MISSING_TYPE, MISSING
from typing import (Any, Callable, Dict, Iterable, List, Optional, Set, Tuple,
                    Type, Union, overload)

from simple_parsing.utils import (Dataclass, K, SimpleValueType, T, V,
                                  get_type_arguments, is_union)

from ..logging_utils import get_logger

logger = get_logger(__file__)

def field(default: Union[T, _MISSING_TYPE] = MISSING,
          alias: Optional[Union[str, List[str]]] = None,
          *,
          to_dict: bool=True,
          encoding_fn: Callable[[T], Any]=None,
          decoding_fn: Callable[[Any], T]=None,
          # dataclasses.field arguments
          default_factory: Union[Callable[[], T], _MISSING_TYPE] = MISSING,
          init: bool = True,
          repr: bool = True,
          hash: Optional[bool] = None,
          compare: bool = True,
          metadata: Optional[Dict[str, Any]] = None,
          **custom_argparse_args: Any) -> T:
    """Calls the `dataclasses.field` function, and leftover arguments are fed
    directly to the `ArgumentParser.add_argument(*option_strings, **kwargs)`
    method.

    Parameters
    ----------
    default : Union[T, _MISSING_TYPE], optional
        The default field value (same as in `dataclasses.field`), by default MISSING
    alias : Union[str, List[str]], optional
        Additional option_strings to pass to the `add_argument` method, by
        default None. When passing strings which do not start by "-" or "--", 
        will be prefixed with "-" if the string is one character and by "--"
        otherwise.
    
    ## Serialization-related Keyword Arguments:

    to_dict : bool
        Wether to include this field in the dictionary when calling `to_dict()`.
        Defaults to True.
        Only has an effect when the dataclass containing this field is
        `Serializable`.
    encoding_fn : Callable[[T], Any], optional
        Function to apply to this field's value when encoding the dataclass to a
        dict. Only has an effect when the dataclass containing this field is
        `Serializable`.
    decoding_fn : Callable[[Any], T]. optional
        Function to use in order to recover a the value of this field from a
        serialized entry in a dictionary (inside `cls.from_dict`).
        Only has an effect when the dataclass containing this field is
        `Serializable`.
    
    ## Keyword Arguments of `dataclasses.field`

    default_factory : Union[Callable[[], T], _MISSING_TYPE], optional
        (same as in `dataclasses.field`), by default None
    init : bool, optional
        (same as in `dataclasses.field`), by default True
    repr : bool, optional
        (same as in `dataclasses.field`), by default True
    hash : bool, optional
        (same as in `dataclasses.field`), by default None
    compare : bool, optional
        (same as in `dataclasses.field`), by default True
    metadata : Dict[str, Any], optional
        (same as in `dataclasses.field`), by default None

    Returns
    -------
    T
        The value returned by the `dataclasses.field` function.
    """
    _metadata: Dict[str, Any] = metadata if metadata is not None else {}
    if alias:
        _metadata["alias"] = alias if isinstance(alias, list) else [alias]
    _metadata.update(dict(to_dict=to_dict))
    if encoding_fn is not None:
        _metadata.update(dict(encoding_fn=encoding_fn))
    if decoding_fn is not None:
        _metadata.update(dict(decoding_fn=decoding_fn))

    if custom_argparse_args:
        _metadata.update({"custom_args": custom_argparse_args})
        
        action = custom_argparse_args.get("action")
        if action == "store_false":
            if default not in {MISSING, True}:
                raise RuntimeError("default should either not be passed or set "
                                   "to True when using the store_false action.")
            default = True  # type: ignore
        
        elif action == "store_true":
            if default not in {MISSING, False}:
                raise RuntimeError("default should either not be passed or set "
                                   "to False when using the store_true action.")
            default = False  # type: ignore

    if default is not MISSING:
        return dataclasses.field(  # type: ignore
            default=default,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )
    elif not isinstance(default_factory, dataclasses._MISSING_TYPE):
        return dataclasses.field(
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )
    else:
        return dataclasses.field(
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata
        )


def choice(*choices: T, default: T = None, **kwargs: Any) -> T:
    """ Makes a regular attribute, whose value, when parsed from the 
    command-line, can only be one contained in `choices`, with a default value 
    of `default`.

    Returns a regular `dataclasses.field()`, but with metadata which indicates  
    the allowed values.

    (New:) If `choices` is a dictionary, then passing the 'key' will result in
    the corresponding value being used. The values may be objects, for example.

    Args:
        default (T, optional): The default value of the field. Defaults to None,
        in which case the command-line argument is required.

    Raises:
        ValueError: If the default value isn't part of the given choices.

    Returns:
        T: the result of the usual `dataclasses.field()` function (a dataclass field/attribute).
    """
    assert len(choices) > 0, "Choice requires at least one positional argument!"
    if isinstance(choices[0], dict):
        if len(choices) > 1:
            raise ValueError(f"'choices' should be either a list of values or "
                             f"a single dictionary. (Received {choices})")
        choice_dict = choices[0]

        # if the choices is a dict, the options are the keys
        choices = tuple(choice_dict.keys())

        # save the info about the choice_dict in the field metadata.
        metadata = kwargs.setdefault("metadata", {})
        metadata["choice_dict"] = choice_dict
        if default is not None and default in choice_dict:
            return field(default_factory=functools.partial(choice_dict.get, default), choices=choices, **kwargs)  # type: ignore


    if default is not None and default not in choices:
        raise ValueError(f"Default value of {default} is not a valid option! "
                         f"(options: {choices})")
    
    return field(default=default, choices=choices, **kwargs)  # type: ignore



def list_field(*default_items: T, **kwargs) -> List[T]:
    """shorthand function for setting a `list` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same list.

    Accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        List[T]: a `dataclasses.field` of type `list`, containing the `default_items`. 
    """
    default = kwargs.pop("default", None)
    if isinstance(default, list):
        # can't have that. field wants a default_factory.
        # we just give back a copy of the list as a default factory,
        # but this should be discouraged.
        from copy import deepcopy
        kwargs["default_factory"] = lambda: deepcopy(default)
    return mutable_field(list, default_items, **kwargs)


def dict_field(default_items: Union[Dict[K, V], Iterable[Tuple[K, V]]]=None, **kwargs) -> Dict[K, V]:
    """shorthand function for setting a `dict` attribute on a dataclass,
    so that every instance of the dataclass doesn't share the same `dict`.

    NOTE: Do not use keyword arguments as you usually would with a dictionary
    (as in something like `dict_field(a=1, b=2, c=3)`). Instead pass in a
    dictionary instance with the items: `dict_field(dict(a=1, b=2, c=3))`.
    The reason for this is that the keyword arguments are interpreted as custom
    argparse arguments, rather than arguments of the `dict` function!) 

    Also accepts any of the arguments of the `dataclasses.field` function.

    Returns:
        Dict[K, V]: a `dataclasses.Field` of type `Dict[K, V]`, containing the `default_items`. 
    """
    if default_items is None:
        default_items = {}
    elif isinstance(default_items, dict):
        default_items = default_items.items()
    return mutable_field(dict, default_items, **kwargs)


def set_field(*default_items: T, **kwargs) -> Set[T]:
    return mutable_field(set, default_items, **kwargs)


def mutable_field(_type: Type[T], *args, init: bool = True, repr: bool = True, hash: bool = None, compare: bool = True, metadata: Dict[str, Any] = None, **kwargs) -> T:
    # TODO: Check wether some of the keyword arguments are destined for the `field` function, or for the partial?    
    default_factory = kwargs.pop("default_factory", functools.partial(_type, *args))
    return field(default_factory=default_factory, init=init, repr=repr, hash=hash, compare=compare, metadata=metadata, **kwargs)


MutableField = mutable_field


class SimpleHelpFormatter(argparse.ArgumentDefaultsHelpFormatter,
                          argparse.MetavarTypeHelpFormatter,
                          argparse.RawDescriptionHelpFormatter):
    """Little shorthand for using some useful HelpFormatters from argparse.
    
    This class inherits from argparse's `ArgumentDefaultHelpFormatter`,
    `MetavarTypeHelpFormatter` and `RawDescriptionHelpFormatter` classes.

    This produces the following resulting actions:
    - adds a "(default: xyz)" for each argument with a default
    - uses the name of the argument type as the metavar. For example, gives
      "-n int" instead of "-n N" in the usage and description of the arguments.
    - Conserves the formatting of the class and argument docstrings, if given.
    """

    def _get_default_metavar_for_optional(self, action):
        try:
            return super()._get_default_metavar_for_optional(action)
        except:
            return self._get_metavar_for_type(action.type, optional=True)


    def _get_default_metavar_for_positional(self, action):
        try:
            return super()._get_default_metavar_for_positional(action)
        except:
            return self._get_metavar_for_type(action.type, optional=False)

    def _get_metavar_for_type(self, t: Type, optional: bool=False) -> str:
        if hasattr(t, "__name__"):
            return t.__name__
        elif is_union(t):
            type_args = list(get_type_arguments(t))
            
            none_type = type(None)
            while none_type in type_args:  # type: ignore
                type_args.remove(none_type)  # type: ignore
            
            string = ""
            if optional:
                string += "["
            middle = []
            for t_ in type_args:
                middle.append(self._get_metavar_for_type(t_, optional=optional))
            string += "|".join(middle)
            if optional:
                string += "]"
            return string
        else:
            return str(t)

Formatter = SimpleHelpFormatter

def subparsers(subcommands: Dict[str, Type[Dataclass]]) -> Any:
    return field(metadata={
        "subparsers": subcommands,
    })



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from . import helpers, utils, wrappers
from .conflicts import ConflictResolution
from .helpers import (MutableField, Serializable, SimpleHelpFormatter, choice,
                      field, list_field, mutable_field, subparsers)
from .parsing import ArgumentParser
from .utils import InconsistentArgumentError

__all__ = [
    "helpers", "utils", "wrappers",
    "ConflictResolution",
    "MutableField", "Serializable", "SimpleHelpFormatter", "choice",
    "field", "list_field", "mutable_field", "subparsers",
    "ArgumentParser",
    "InconsistentArgumentError",
]
